import json
import re
from test import conftest

import msgspec
import pytest
from pydantic import ValidationError

import turvallisuusneuvonta.csaf.product as product
import turvallisuusneuvonta.csaf.vulnerability as vulnerability
from turvallisuusneuvonta.csaf.cvss.cvss import CVSS31

# , SeverityType, Version as CvssVersion

CVSS31_BASE_SEVERITY_LOG4J = 'CRITICAL'  # str(SeverityType.critical)


def test_vulnerability_empty():
    assert isinstance(vulnerability.Vulnerability(), vulnerability.Vulnerability)


def test_vulnerability_text():
    message = 'BaseModel.__init__() takes 1 positional argument but 3 were given'
    with pytest.raises(TypeError, match=re.escape(message)):
        _ = vulnerability.Vulnerability('some', 'text')  # type: ignore


def test_vulnerability_score_text():
    message = 'BaseModel.__init__() takes 1 positional argument but 2 were given'
    with pytest.raises(TypeError, match=re.escape(message)):
        _ = vulnerability.Score('text')  # type: ignore


def test_vulnerability_score_cvss31_log4j_cve_2021_44228():
    data = {
        'version': '3.1',  # str(CvssVersion.three_wun),
        'vectorString': conftest.CVSS31_VECTOR_STRING_LOG4J,
        'baseScore': conftest.CVSS31_BASE_SCORE_LOG4J,
        'baseSeverity': CVSS31_BASE_SEVERITY_LOG4J,
    }
    as_json = json.dumps(data)
    cvss_cve_2021_44228 = CVSS31.model_validate_json(as_json)
    message = '1 validation error for Score'
    with pytest.raises(ValidationError, match=message) as err:
        _ = vulnerability.Score(cvss_v3=cvss_cve_2021_44228, products=[])  # type: ignore
    assert '\nproducts\n  List should have at least 1 item after validation, not 0' in str(err.value)


def test_vulnerability_score_cvss31_log4j_cve_2021_44228_product_foo():
    data = {
        'version': '3.1',  # str(CvssVersion.three_wun),
        'vectorString': conftest.CVSS31_VECTOR_STRING_LOG4J,
        'baseScore': conftest.CVSS31_BASE_SCORE_LOG4J,
        'baseSeverity': CVSS31_BASE_SEVERITY_LOG4J,
    }
    # pr_ref = product.ReferenceTokenForProductInstance(value='log4j-123')
    products = product.Products(['log4j-123'])
    as_json = json.dumps(data)
    cvss_cve_2021_44228 = CVSS31.model_validate_json(as_json)
    score = vulnerability.Score(cvss_v3=cvss_cve_2021_44228, products=products)
    assert score.cvss_v3 == cvss_cve_2021_44228
    assert score.products == products


def test_vulnerability_score_loads_cvss31_log4j_cve_2021_44228_product_foo():
    score = vulnerability.Score.model_validate_json(json.dumps(conftest.VULNERABILITY_SCORE_LOG4J))
    assert msgspec.json.decode(score.model_dump_json()) == conftest.VULNERABILITY_SCORE_LOG4J


def test_vulnerability_cwe_text():
    message = 'BaseModel.__init__() takes 1 positional argument but 2 were given'
    with pytest.raises(TypeError, match=re.escape(message)):
        _ = vulnerability.Cwe('text')  # type: ignore


def test_vulnerability_cwe_352():
    data = {
        'id': conftest.CWE_ID_352,
        'name': conftest.CWE_NAME_352,
    }
    cwe_352 = vulnerability.Cwe(**data)
    assert cwe_352.id == conftest.CWE_ID_352
    assert cwe_352.name == conftest.CWE_NAME_352


def test_vulnerability_entitlement_none():
    message = '1 validation error for Entitlement'
    with pytest.raises(ValidationError, match=message) as err:
        _ = vulnerability.Entitlement()  # type: ignore
    assert 'Input should be a valid string' in str(err.value)


def test_vulnerability_entitlement_empty():
    message = '1 validation error for Entitlement'
    with pytest.raises(ValidationError, match=message) as err:
        _ = vulnerability.Entitlement('')
    assert 'String should have at least 1 character' in str(err.value)


def test_vulnerability_entitlement_with_value():
    value = 'All people on this planet'
    all_people_entitled = vulnerability.Entitlement(value)
    assert all_people_entitled.root == value


def test_vulnerability_restart_req_by_rem_none():
    message = '1 validation error for RestartRequired'
    with pytest.raises(ValidationError, match=message) as err:
        _ = vulnerability.RestartRequired()  # type: ignore
    assert '\ncategory\n  Field required' in str(err.value)


def test_vulnerability_restart_req_by_rem_empty_category():
    message = '1 validation error for RestartRequired'
    with pytest.raises(ValidationError, match=message) as err:
        _ = vulnerability.RestartRequired(category='')  # type: ignore
    expected = (
        "\ncategory\n  Input should be 'connected', 'dependencies', 'machine', 'none',"
        " 'parent', 'service', 'system', 'vulnerable_component' or 'zone'"
    )
    assert expected in str(err.value)


def test_vulnerability_restart_req_by_rem_valid_category():
    category = vulnerability.RestartRequiredCategory.none
    restart_req_none = vulnerability.RestartRequired(category=category)
    assert restart_req_none.category == category


def test_vulnerability_restart_req_by_rem_valid_category_details_empty():
    category = vulnerability.RestartRequiredCategory.none
    message = '1 validation error for RestartRequired'
    with pytest.raises(ValidationError, match=message) as err:
        _ = vulnerability.RestartRequired(category=category, details='')
    assert '\ndetails\n  String should have at least 1 character' in str(err.value)


def test_vulnerability_restart_req_by_rem_valid_category_details_ok():
    category = vulnerability.RestartRequiredCategory.none
    details = 'And that is all I can say about that.'
    restart_req_none = vulnerability.RestartRequired(category=category, details=details)
    assert restart_req_none.category == category
    assert restart_req_none.details == details


def test_vulnerability_remediation_none():
    message = '2 validation errors for Remediation'
    with pytest.raises(ValidationError, match=message) as err:
        _ = vulnerability.Remediation()
    assert '\ncategory\n  Field required' in str(err.value)
    assert '\ndetails\n  Field required' in str(err.value)


def test_vulnerability_remediation_empty_category():
    message = '2 validation errors for Remediation'
    with pytest.raises(ValidationError, match=message) as err:
        _ = vulnerability.Remediation(category='')  # type: ignore
    expected = (
        "\ncategory\n  Input should be 'mitigation', 'no_fix_planned', 'none_available', 'vendor_fix' or 'workaround'"
    )
    assert expected in str(err.value)
    assert '\ndetails\n  Field required' in str(err.value)


def test_vulnerability_remediation_valid_category_details_empty():
    category = vulnerability.RemediationCategory.none_available
    message = '1 validation error for Remediation'
    with pytest.raises(ValidationError, match=message) as err:
        _ = vulnerability.Remediation(category=category, details='')
    assert '\ndetails\n  String should have at least 1 character' in str(err.value)


def test_vulnerability_remediation_req_by_rem_valid_category_details_ok():
    category = vulnerability.RemediationCategory.none_available
    details = 'And that is all I can say about that.'
    remediation_none = vulnerability.Remediation(category=category, details=details)
    assert remediation_none.category == category
    assert remediation_none.details == details


def test_vulnerability_threat_none():
    message = '2 validation errors for Threat'
    with pytest.raises(ValidationError, match=message) as err:
        _ = vulnerability.Threat()
    assert '\ncategory\n  Field required' in str(err.value)
    assert '\ndetails\n  Field required' in str(err.value)


def test_vulnerability_threat_empty_category():
    message = '2 validation errors for Threat'
    with pytest.raises(ValidationError, match=message) as err:
        _ = vulnerability.Threat(category='')  # type: ignore
    expected = "\ncategory\n  Input should be 'exploit_status', 'impact' or 'target_set'"
    assert expected in str(err.value)
    assert '\ndetails\n  Field required' in str(err.value)


def test_vulnerability_threat_valid_category_details_empty():
    category = vulnerability.ThreatCategory.exploit_status
    message = '1 validation error for Threat'
    with pytest.raises(ValidationError, match=message) as err:
        _ = vulnerability.Threat(category=category, details='')
    assert '\ndetails\n  String should have at least 1 character' in str(err.value)


def test_vulnerability_threat_req_by_rem_valid_category_details_ok():
    category = vulnerability.ThreatCategory.exploit_status
    details = 'And that is all I can say about that.'
    threat_exploit = vulnerability.Threat(category=category, details=details)
    assert threat_exploit.category == category
    assert threat_exploit.date is None
    assert threat_exploit.details == details
    assert threat_exploit.group_ids is None
    assert threat_exploit.product_ids is None
